/*	Main.cpp

	Copyright (C) 2004, Tom Mulgrew (tmulgrew@slingshot.co.nz)

	Basic4GL command line compiler build.

	Usage:
		B4GL programname.gb

	Note:
		May be useful for those who wish to build the Basic4GL source on Windows
		but don't have access to Borland C++ Builder.

		This program should run all Basic4GL programs that the official released
		version of Basic4GL runs.

		There is no user interface in this version however.
*/

#include "TomComp.h"
#include "glTextGrid.h"
#include "TomOpenGLBasicLib.h"
#include "glBasicLib_gl.h"
#include "glbasiclib_glu.h" 
#include "TomStdBasicLib.h"
#include "TomWindowsBasicLib.h"
#include "TomTextBasicLib.h"
#include "TomJoystickBasicLib.h"
#include "TomTrigBasicLib.h"
#include "TomFileIOBasicLib.h"
#include "glSpriteEngine.h"
#include "EmbeddedFiles.h"
#include "Standalone.h"

#ifdef _DEBUG
#define FULLSCREEN false
#else 
#define FULLSCREEN true
#endif

char buffer [4096];

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{

	// Allocate objects
    TomVM               vm;									// Virtual machine		
    TomBasicCompiler    comp (vm);							// Compiler
	FileOpener			files;

	// Load program into compiler
	std::ifstream file (lpCmdLine);
	if (file.fail ()) {
		MessageBox (NULL, "Failed to open input file\r\nUsage: B4GL filename.gb", "Error opening file", MB_OK);
		return 0;
	}
	while (!file.eof ()) {
		file.getline (buffer, 4095);
		comp.Parser ().SourceCode ().push_back ((std::string) buffer);
	}

	// Create OpenGL window	
    glTextGridWindow	glWin (	FULLSCREEN,					// Fullscreen mode
								true,						// Frame around window (windowed mode only)
								640,						// Width
								480,						// Height
								0,							// Bpp (fullscreen mode only, 0 = same as desktop, otherwise 16 or 32)
								lpCmdLine,					// Window title
								RGM_RECREATEWINDOW);		// Not used. How to REcreate OpenGL context. But we're only initialising OpenGL once before exiting, so not important.

	// Initialise text and sprite engine
    glSpriteEngine		glText("charset.png", &files, 25, 40, 16, 16);	// Setup a text grid for the window
    glWin.SetTextGrid (&glText);									
	if (glText.Error ()) {
		MessageBox (glWin.GetHWND (), glText.GetError ().c_str (), "ERROR!", MB_OK);
		return 0;
	}

	// Initialise function libraries
    InitTomStdBasicLib (comp);							// Standard library
    InitTomWindowsBasicLib (comp, &files);				// Windows specific library
    InitTomOpenGLBasicLib (comp, &glWin, &files);		// OpenGL
    InitTomTextBasicLib (comp, &glWin, &glText);
    InitGLBasicLib_gl (comp);
    InitGLBasicLib_glu (comp);
	InitTomJoystickBasicLib (comp, &glWin);
    InitTomTrigBasicLib (comp);							// Trigonometry library
    InitTomFileIOBasicLib (comp, &files);				// File I/O library

	// Compile program
	comp.ClearError ();
	comp.Compile ();
    if (comp.Error ()) {
		MessageBox (glWin.GetHWND (), ((std::string) "Compile error: " + comp.GetError () + ", line " + IntToString (comp.Line () + 1) + ", col " + IntToString (comp.Col () + 1)).c_str (), "ERROR!", MB_OK);
		return 0;
	}
	
	// Activate window
	glWin.Activate ();
	glWin.OpenGLDefaults ();

	// Reset virtual machine
	vm.Reset ();
	try {
		do {

			// Run VM for a number of steps
			vm.Continue (TC_STEPSBETWEENREFRESH);

			// Process windows messages to keep application responsive
			glWin.ProcessWindowsMessages ();
		} while (!vm.Error () && !vm.Done () && !glWin.Closing ());
	}
	catch (...) {
		vm.MiscError ("An exception occured");
	}

	if (vm.Error ()) {

		// Runtime program error

		// Switch back from fullscreen mode (if fullscreen)
		if (glWin.FullScreen ())
			glWin.Hide ();

		// Get error position
		int line, col;
		vm.GetIPInSourceCode (line, col);

		// Display runtime error
		MessageBox (glWin.GetHWND (), ((std::string) vm.GetError () + ", line " + IntToString (line + 1) + ", col " + IntToString (col + 1)).c_str (), "Runtime error", MB_OK);
	}
	else if (vm.Done ()) {

		// Wait for a keypress (or explicit window close)
		glWin.ProcessWindowsMessages(); 
		while (glWin.GetKey () != 0 && !glWin.Closing ()) glWin.ProcessWindowsMessages();
		while (glWin.GetKey () == 0 && !glWin.Closing ()) glWin.ProcessWindowsMessages();
	}

	// Finished
	return 0;
}