//---------------------------------------------------------------------------
// Created 24-Apr-2004: Thomas Mulgrew (tmulgrew@slingshot.co.nz)
// Copyright (C) Thomas Mulgrew

// Simple Winsock based TCP/IP library

#pragma hdrstop

#include "TomNetBasicLib.h"
#include <winsock2.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

////////////////////////////////////////////////////////////////////////////////
// NetSocket
//
// A network socket (wrapper for Winsock SOCKET type)
class NetSocket : public HasErrorState {
protected:
    SOCKET  m_socket;           // Winsock SOCKET
    bool    m_active;           // True if socket is active

    void SocketDefaults ();

public:
    NetSocket ();
    virtual ~NetSocket ();

    SOCKET  Socket () { return m_socket; }
    bool    Active () { return m_active; }

    void    Close ();
};

NetSocket::NetSocket () {
    m_active = false;
}

NetSocket::~NetSocket () {
    Close ();
}

void NetSocket::Close () {
    if (m_active) {
    	shutdown (m_socket, SD_SEND);
	    closesocket (m_socket);
        m_active = false;
    }
}

void NetSocket::SocketDefaults () {
//    bool noDelay = true;
//    setsockopt (m_socket, IPPROTO_TCP, TCP_NODELAY, (char *) &noDelay, sizeof (noDelay));
}

////////////////////////////////////////////////////////////////////////////////
// NetServer
//
// A TCP/IP network server.
class NetServer : public NetSocket {
public:
    NetServer (unsigned short port, int backlog);
};

NetServer::NetServer (unsigned short port, int backlog) {
	m_socket = socket (AF_INET, SOCK_STREAM, 0);
    SocketDefaults ();
//    SetPriorityClass (GetModuleHandle (NULL), HIGH_PRIORITY_CLASS);

	// Bind to requested port
	sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons (port);
	addr.sin_addr.s_addr = htonl (INADDR_ANY);
	if (bind (m_socket, (LPSOCKADDR)&addr, sizeof (addr)) == SOCKET_ERROR) {
        SetError ((std::string) "Error binding to port " + IntToString (port));
        return;
	}

	// Start listening for connections
	if (listen (m_socket, backlog) == SOCKET_ERROR) {
        SetError ((std::string) "Error listening on port " + IntToString (port));
        return;
	}

    // Server created!
    m_active = true;
}

////////////////////////////////////////////////////////////////////////////////
// NetConnection
//
// A TCP/IP network connection.
// (Wrapper for Winsock SOCKET.)
class NetConnection : public NetSocket {
public:
    NetConnection (std::string address, unsigned short port);       // Connect to address
    NetConnection (NetServer& server);                              // Accept a connection from the server

    bool DataWaiting ();
    bool Read (void *ptr, int len);
    void Write (void *ptr, int len);
};

NetConnection::NetConnection (std::string address, unsigned short port) {

    // Create socket handle
    m_socket = socket (AF_INET, SOCK_STREAM, 0);
    SocketDefaults ();

    // Find target address
  	sockaddr_in target;
	target.sin_family = AF_INET;
	target.sin_port = htons (port);

    // Use numeric form for localhost
    if (LowerCase (address) == "localhost")
        address = "127.0.0.1";

    // Try numeric address first
    target.sin_addr.s_addr = inet_addr (address.c_str ());
    if (target.sin_addr.s_addr == INADDR_NONE) {

        // Numeric failed. Perform DNS lookup and try again.
        hostent *he = gethostbyname (address.c_str ());
        if (he == NULL) {
            SetError ("Unable to resolve address: " + address);
            return;
        }
        target.sin_addr.s_addr = *((u_long*)he->h_addr_list[0]);
    }

    // Connect
	if (connect (m_socket, (sockaddr *) &target, sizeof (target)) == SOCKET_ERROR) {
		SetError ("Error connecting to: " + address);
        return;
	}

    // Connection created
    m_active = true;
}

NetConnection::NetConnection (NetServer& server) {

    // Accept a connection from the server.
    // This is a non-blocking function, so we must first ensure that there is
    // a connection waiting to be accepted.
    fd_set read, write, except;
    FD_ZERO (&read);
    FD_ZERO (&write);
    FD_ZERO (&except);
    FD_SET (server.Socket (), &read);
    timeval timeout = {0, 0};                   // No timeout

    // Check that a connection is waiting
    select (0, &read, &write, &except, &timeout);
    if (!FD_ISSET (server.Socket (), &read))
        return;
    
    // Accept the waiting connection
    m_socket = accept (server.Socket (), NULL, NULL);
    if (m_socket == INVALID_SOCKET) {
        SetError ("Error accepting connection");
        return;
    }
    SocketDefaults ();

    // Connection accepted
    m_active = true;
}

bool NetConnection::DataWaiting () {
    if (!m_active)
        return false;

    // Look for data
    fd_set read, write, except;
    FD_ZERO (&read);
    FD_ZERO (&write);
    FD_ZERO (&except);
    FD_SET (m_socket, &read);
    timeval timeout = {0, 0};                  

    // Check that data is waiting
    select (0, &read, &write, &except, &timeout);
    return FD_ISSET (m_socket, &read);
}

bool NetConnection::Read (void *ptr, int len) {
    assert (m_active);
    assert (ptr != NULL);
    assert (len >= 0);

    int received = recv (m_socket, (char *) ptr, len, 0);
    if (received == 0 || received == SOCKET_ERROR) {

        // Disconnect received
        Close ();
        return false;
    }
    return received == len;
}

void NetConnection::Write (void *ptr, int len) {
    assert (m_active);
    assert (ptr != NULL);
    assert (len >= 0);

    send (m_socket, (char *) ptr, len, 0);
}

////////////////////////////////////////////////////////////////////////////////
// Network resource stores
typedef vmPointerResourceStore<NetServer>       NetServerResourceStore;
typedef vmPointerResourceStore<NetConnection>   NetConnectionResourceStore;

// Globals
NetServerResourceStore      servers;
NetConnectionResourceStore  connections;
std::string                 netError = "";
bool                        winsockInitialised = false;

////////////////////////////////////////////////////////////////////////////////
// Pre-run initialisation
static void Init (TomVM& vm) {
    netError = "";
//    SetPriorityClass (GetModuleHandle (NULL), NORMAL_PRIORITY_CLASS);
}

////////////////////////////////////////////////////////////////////////////////
// Function wrappers

bool CheckWinsock () {
    if (!winsockInitialised) {
    	WSADATA wsa;
	    if (WSAStartup (0x0202, &wsa) == 0)
            winsockInitialised = true;
        else
            netError = "Winsock initialisation failure";
    }
    return winsockInitialised;
}

NetConnection *GetConnection (TomVM& vm, int paramIndex) {
    if (!CheckWinsock ())
        return NULL;
    int handle = vm.GetIntParam (paramIndex);
    if (!connections.IndexStored (handle)) {
        netError = "Invalid net connection handle";
        return NULL;
    }
    return connections.Value (handle);
}

NetServer *GetServer (TomVM& vm, int paramIndex) {
    if (!CheckWinsock ())
        return NULL;
    int handle = vm.GetIntParam (paramIndex);
    if (!servers.IndexStored (handle)) {
        netError = "Invalid net server handle";
        return NULL;
    }
    return servers.Value (handle);
}

void WrapNetError (TomVM& vm) {
    vm.RegString () = netError;
}

inline void CheckError (NetSocket& socket) {
    netError = socket.Error () ? socket.GetError () : (std::string) "";
}

void WrapNewNetServer (TomVM& vm) {
    if (!CheckWinsock ()) {
        vm.Reg ().IntVal () = 0;
        return;
    }
    unsigned short port = vm.GetIntParam (2);
    int backlog = vm.GetIntParam (1);
    NetServer *server = new NetServer (port, backlog);
    if (server->Active ())
        vm.Reg ().IntVal () = servers.Alloc (server);
    else {
        CheckError (*server);
        delete server;
        vm.Reg ().IntVal () = 0;
    }
}

void WrapNewNetCon (TomVM& vm) {
    if (!CheckWinsock ()) {
        vm.Reg ().IntVal () = 0;
        return;
    }
    std::string address = vm.GetStringParam (2);
    unsigned short port = vm.GetIntParam (1);
    NetConnection *connection = new NetConnection (address, port);
    if (connection->Active ())
        vm.Reg ().IntVal () = connections.Alloc (connection);
    else {
        CheckError (*connection);
        delete connection;
        vm.Reg ().IntVal () = 0;
    }
}

void WrapNewNetCon_2 (TomVM& vm) {

    // Find server
    NetServer *server = GetServer (vm, 1);
    if (server == NULL) {
        vm.Reg ().IntVal () = 0;
        return;
    }

    // Create connection from server
    // Note:
    //  If no connection is available, wrapper will return 0, but no error will
    //  be set.    
    NetConnection *connection = new NetConnection (*server);
    if (connection->Active ())
        vm.Reg ().IntVal () = connections.Alloc (connection);
    else {
        CheckError (*connection);
        delete connection;
        vm.Reg ().IntVal () = 0;
    }
}

void WrapNetConActive (TomVM& vm) {

    // Find connection
    NetConnection *connection = GetConnection (vm, 1);
    if (connection == NULL)
        vm.Reg ().IntVal () = 0;
    else
        vm.Reg ().IntVal () = connection->Active () ? -1 : 0;
}

void WrapNetServerActive (TomVM& vm) {

    // Find server
    NetServer *server = GetServer (vm, 1);
    if (server == NULL)
        vm.Reg ().IntVal () = 0;
    else
        vm.Reg ().IntVal () = server->Active () ? -1 : 0;
}

void WrapNetDataWaiting (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    vm.Reg ().IntVal () =   connection != NULL
                        &&  connection->Active ()
                        &&  connection->DataWaiting ();
}

void WrapNetReadInt (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        int data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().IntVal () = ntohl (data);
            return;
        }
    }
    vm.Reg ().IntVal () = 0;
}

void WrapNetReadChar (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        char data;
        if (connection->Read (&data, sizeof (data))) {
            vm.RegString () = data;
            return;
        }
    }
    vm.RegString () = "";
}

void WrapNetReadByte (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        byte data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().IntVal () = data;
            return;
        }
    }
    vm.Reg ().IntVal () = 0;
}

void WrapNetReadWord (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        WORD data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().IntVal () = ntohs (data);
            return;
        }
    }
    vm.Reg ().IntVal () = 0;
}

void WrapNetReadFloat (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        float data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().RealVal () = data;
            return;
        }
    }
    vm.Reg ().RealVal () = 0;
}

void WrapNetReadReal (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        vmReal data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().RealVal () = data;
            return;
        }
    }
    vm.Reg ().RealVal () = 0;
}

void WrapNetReadDouble (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 1);
    if (    connection != NULL
        &&  connection->Active ()
        &&  connection->DataWaiting ()) {
        double data;
        if (connection->Read (&data, sizeof (data))) {
            vm.Reg ().RealVal () = data;
            return;
        }
    }
    vm.Reg ().RealVal () = 0;
}

void WrapNetWriteInt (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        int data = htonl (vm.GetIntParam (1));
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteChar (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        char data = vm.GetStringParam (1) [0];
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteString (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        std::string data = vm.GetStringParam (1);
        connection->Write ((void *) data.c_str (), data.length ());
    }
}

void WrapNetWriteByte (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        byte data = vm.GetIntParam (1);
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteWord (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        WORD data = htons (vm.GetIntParam (1));
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteFloat (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        float data = vm.GetRealParam (1);
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteReal (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        vmReal data = vm.GetRealParam (1);
        connection->Write (&data, sizeof (data));
    }
}

void WrapNetWriteDouble (TomVM& vm) {
    NetConnection *connection = GetConnection (vm, 2);
    if (connection != NULL && connection->Active ()) {
        double data = vm.GetRealParam (1);
        connection->Write (&data, sizeof (data));
    }
}

void WrapCloseNetCon (TomVM& vm) {
    int handle = vm.GetIntParam (1);
    if (connections.IndexStored (handle))
        connections.Free (handle);
}

void WrapCloseNetServer (TomVM& vm) {
    int handle = vm.GetIntParam (1);
    if (servers.IndexStored (handle))
        servers.Free (handle);
}

void InitTomNetBasicLib (TomBasicCompiler& comp) {

    // Register resources
    comp.VM ().AddResources (servers);
    comp.VM ().AddResources (connections);

    // Register initialisation func
    comp.VM().AddInitFunc (Init);

    // Register functions
    comp.AddFunction ("NetError",           WrapNetError,           compParamTypeList (),                           true, true, VTP_STRING);
    comp.AddFunction ("NewNetServer",       WrapNewNetServer,       compParamTypeList () << VTP_INT << VTP_INT,     true, true, VTP_INT, true);
    comp.AddFunction ("NewNetCon",          WrapNewNetCon_2,        compParamTypeList () << VTP_INT,                true, true, VTP_INT, true);
    comp.AddFunction ("NewNetCon",          WrapNewNetCon,          compParamTypeList () << VTP_STRING << VTP_INT,  true, true, VTP_INT, true);
    comp.AddFunction ("NetConActive",       WrapNetConActive,       compParamTypeList () << VTP_INT,                true, true, VTP_INT);
    comp.AddFunction ("NetServerActive",    WrapNetServerActive,    compParamTypeList () << VTP_INT,                true, true, VTP_INT);
    comp.AddFunction ("NetDataWaiting",     WrapNetDataWaiting,     compParamTypeList () << VTP_INT,                true, true, VTP_INT);

    comp.AddFunction ("NetReadInt",         WrapNetReadInt,         compParamTypeList () << VTP_INT,                true, true, VTP_INT);
    comp.AddFunction ("NetReadChar",        WrapNetReadChar,        compParamTypeList () << VTP_INT,                true, true, VTP_STRING);
    comp.AddFunction ("NetReadByte",        WrapNetReadByte,        compParamTypeList () << VTP_INT,                true, true, VTP_INT);
    comp.AddFunction ("NetReadWord",        WrapNetReadWord,        compParamTypeList () << VTP_INT,                true, true, VTP_INT);
    comp.AddFunction ("NetReadFloat",       WrapNetReadFloat,       compParamTypeList () << VTP_INT,                true, true, VTP_REAL);
    comp.AddFunction ("NetReadReal",        WrapNetReadReal,        compParamTypeList () << VTP_INT,                true, true, VTP_REAL);
    comp.AddFunction ("NetReadDouble",      WrapNetReadDouble,      compParamTypeList () << VTP_INT,                true, true, VTP_REAL);

    comp.AddFunction ("NetWriteInt",        WrapNetWriteInt,        compParamTypeList () << VTP_INT << VTP_INT,     true, false, VTP_INT);
    comp.AddFunction ("NetWriteChar",       WrapNetWriteChar,       compParamTypeList () << VTP_INT << VTP_STRING,  true, false, VTP_INT);
    comp.AddFunction ("NetWriteString",     WrapNetWriteString,     compParamTypeList () << VTP_INT << VTP_STRING,  true, false, VTP_INT);
    comp.AddFunction ("NetWriteByte",       WrapNetWriteByte,       compParamTypeList () << VTP_INT << VTP_INT,     true, false, VTP_INT);
    comp.AddFunction ("NetWriteWord",       WrapNetWriteWord,       compParamTypeList () << VTP_INT << VTP_INT,     true, false, VTP_INT);
    comp.AddFunction ("NetWriteFloat",      WrapNetWriteFloat,      compParamTypeList () << VTP_INT << VTP_REAL,    true, false, VTP_INT);
    comp.AddFunction ("NetWriteReal",       WrapNetWriteReal,       compParamTypeList () << VTP_INT << VTP_REAL,    true, false, VTP_INT);
    comp.AddFunction ("NetWriteDouble",     WrapNetWriteDouble,     compParamTypeList () << VTP_INT << VTP_REAL,    true, false, VTP_INT);

    comp.AddFunction ("CloseNetCon",        WrapCloseNetCon,        compParamTypeList () << VTP_INT,                true, false, VTP_INT);
    comp.AddFunction ("CloseNetServer",     WrapCloseNetServer,     compParamTypeList () << VTP_INT,                true, false, VTP_INT);
}

void CleanupTomNetBasicLib () {
    if (winsockInitialised) {
        WSACleanup ();
        winsockInitialised = false;
    }
}
