//---------------------------------------------------------------------------
// Created 27-Sep-2003: Thomas Mulgrew (tmulgrew@slingshot.co.nz)
// Copyright (C) Thomas Mulgrew
#pragma hdrstop

#include "TomTextBasicLib.h"                                       
#include <gl/gl.h>
#include "glSpriteEngine.h"

//---------------------------------------------------------------------------

#ifndef _MSC_VER
#pragma package(smart_init)
#endif

// Global variables
static glWindow    *appWindow;
static glTextGrid  *appText;

////////////////////////////////////////////////////////////////////////////////
// glSpriteStore
//
// A store of glSprites
class glSpriteStore : public vmResourceStore<glBasicSprite *> {
protected:
    virtual void DeleteElement (int index);
public:
    glSpriteStore () : vmResourceStore<glBasicSprite *> (NULL) { ; }
};

void glSpriteStore::DeleteElement (int index) {
    delete Value (index);
}

glSpriteStore sprites;
GLuint boundSprite;

// Text mode constants/variables
enum TextMode {
    TEXT_SIMPLE,        // (Default). After any print function, text is automatically rendered as follows:
                        //      * Back buffer is cleared
                        //      * Text is rendered
                        //      * Buffers are swapped

    TEXT_BUFFERED,      // Must explicitly call DrawText(), as before:
                        //      * Back buffer is cleared
                        //      * Text is rendered
                        //      * Buffers are swapped

    TEXT_OVERLAID       // Same as TEXT_BUFFERED, except DrawText() only renders the text.
                        // Must explictly clear and swap buffers as necessary.
};

TextMode textMode = TEXT_SIMPLE;

// Pre-run initialisation
static void Init (TomVM& vm) {

    // Text rendering defaults
    textMode = TEXT_SIMPLE;
    appText->Resize (25, 40);
    appText->HideCursor ();
    appText->SetColour (MakeColour (220, 220, 255));
    appText->SetTexture (appText->DefaultTexture ());
    appText->Scroll () = true;
    appWindow->SetDontPaint (false);

    // Sprite engine defaults
    glSpriteEngine *spriteEngine = (glSpriteEngine *) appText;
    spriteEngine->SetDefaults ();
    boundSprite = 0;
}

////////////////////////////////////////////////////////////////////////////////
// Helper functions
inline void ForceDraw () {
    if (textMode == TEXT_SIMPLE || textMode == TEXT_BUFFERED)
        glClear (GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    appText->Draw ();

    if (textMode == TEXT_SIMPLE || textMode == TEXT_BUFFERED)
        appWindow->SwapBuffers ();
}

inline void Redraw () {
    if (textMode == TEXT_SIMPLE)
        ForceDraw ();
}

////////////////////////////////////////////////////////////////////////////////
// Function wrappers

void WrapTextMode (TomVM& vm)       { textMode = (TextMode) vm.GetIntParam (1); }
void WrapCls (TomVM& vm)            { appText->Clear ();  Redraw (); }
void WrapPrint (TomVM& vm)          { appText->Write (vm.GetStringParam (1)); Redraw (); }
void WrapPrintr (TomVM& vm)         { appText->Write (vm.GetStringParam (1)); appText->NewLine (); Redraw (); }
void WrapPrintr_2 (TomVM& vm)       { appText->NewLine (); Redraw (); }
void WrapLocate (TomVM& vm)         { appText->MoveCursor (vm.GetIntParam (2), vm.GetIntParam (1)); }
void WrapSwapBuffers (TomVM& vm)    {
    appWindow->SwapBuffers ();
    appWindow->SetDontPaint (false);
}
void WrapDrawText (TomVM& vm)       { ForceDraw (); }
void WrapResizeText (TomVM& vm) {
    int rows = vm.GetIntParam (1), cols = vm.GetIntParam (2);
    if (rows < 1)   rows = 1;
    if (rows > 500) rows = 500;
    if (cols < 1)   cols = 1;
    if (cols > 500) cols = 500;
    appText->Resize (rows, cols);
    Redraw ();
}
void WrapTextRows (TomVM& vm)   { vm.Reg ().IntVal () = appText->Rows (); }
void WrapTextCols (TomVM& vm)   { vm.Reg ().IntVal () = appText->Cols (); }
void WrapInput (TomVM& vm)      { vm.RegString () = appText->GetString (appWindow); }
void WrapInkey (TomVM& vm)      {
    char key = appWindow->GetKey ();
    if (key != 0)   vm.RegString () = key;
    else            vm.RegString () = "";
}
void WrapInScanKey (TomVM& vm)  {
    vm.Reg ().IntVal () = appWindow->GetScanKey ();
}
void WrapClearKeys (TomVM& vm) {
    appWindow->ClearKeyBuffers ();
}
void WrapShowCursor (TomVM& vm) { appText->ShowCursor (); Redraw (); }
void WrapHideCursor (TomVM& vm) { appText->HideCursor (); Redraw (); }
void WrapKeyDown (TomVM& vm)    {
    std::string s = vm.GetStringParam (1);
    if (s == "")    vm.Reg ().IntVal () = 0;
    else            vm.Reg ().IntVal () = appWindow->KeyDown (s [0]) ? -1 : 0;
}
void WrapScanKeyDown (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (index < 0 || index > 255)   vm.Reg ().IntVal () = 0;
    else                            vm.Reg ().IntVal () = appWindow->KeyDown (index) ? -1 : 0;
}
void WrapCharAt (TomVM& vm) {
    char c = appText->TextAt (vm.GetIntParam (2), vm.GetIntParam (1));
    if (c == 0)     vm.RegString () = "";
    else            vm.RegString () = c;
}
void WrapColour (TomVM& vm) {
    appText->SetColour (MakeColour (vm.GetIntParam (3), vm.GetIntParam (2), vm.GetIntParam (1)));
}
void WrapFont (TomVM& vm) {
    appText->SetTexture (vm.GetIntParam (1));
}
void WrapDefaultFont (TomVM& vm) {
    vm.Reg ().IntVal () = appText->DefaultTexture ();
}
void WrapMouseX (TomVM& vm) {
    vm.Reg ().RealVal () = ((vmReal) appWindow->MouseX ()) / appWindow->Width ();
}
void WrapMouseY (TomVM& vm) {
    vm.Reg ().RealVal () = ((vmReal) appWindow->MouseY ()) / appWindow->Height ();
}
void WrapMouseXD (TomVM& vm) {
    vm.Reg ().RealVal () = ((vmReal) appWindow->MouseXD ()) / appWindow->Width () * 2;
}
void WrapMouseYD (TomVM& vm) {
    vm.Reg ().RealVal () = ((vmReal) appWindow->MouseYD ()) / appWindow->Height () * 2; 
}
void WrapMouseButton (TomVM& vm) {
    int button = vm.GetIntParam (1);
    if (button >= 0 && button <= 2)
        vm.Reg ().IntVal () = appWindow->MouseButton (button) ? -1 : 0;
    else
        vm.Reg ().IntVal () = 0;
}
void WrapMouseWheel (TomVM& vm) {
    vm.Reg ().IntVal () = appWindow->MouseWheel ();
}

// Sprite functions
glSprite *NewSprite (TomVM& vm) {

    // Create sprite object
    glSprite *sprite = new glSprite;

    // Add to sprite engine (so we can draw it)
    ((glSpriteEngine *) appText)->AddSprite(sprite);

    // Add to store (so we can track it), and return index to VM
    boundSprite = sprites.Alloc(sprite);
    vm.Reg ().IntVal () = boundSprite;
    return sprite;
}
glTileMap *NewTileMap (TomVM& vm) {

    // Create tile map object
    glTileMap *tileMap = new glTileMap;

    // Add to sprite engine (so we can draw it)
    ((glSpriteEngine *) appText)->AddSprite(tileMap);

    // Add to store (so we can track it), and return index to VM
    boundSprite = sprites.Alloc(tileMap);
    vm.Reg ().IntVal () = boundSprite;
    return tileMap;
}

void WrapNewSprite (TomVM& vm) {

    // Allocate sprite
    NewSprite (vm);
    Redraw ();
}
void WrapNewSprite_2 (TomVM& vm) {

    // Allocate sprite, and set single texture
    NewSprite (vm)->SetTexture (vm.GetIntParam(1));
    Redraw ();
}
bool GetTextures (TomVM& vm, int paramIndex, TextureVec& dest) {

    // Read in texture array and convert to vector (for storage in sprite)
    GLuint frames [65536];
    int size = ArrayDimensionSize (vm.Data (), vm.GetIntParam (paramIndex), 0);
    if (size < 1 || size > 65536) {
        vm.FunctionError ("Texture array size must be 1-65536");
        return false;
    }
    ReadArray (vm.Data (), vm.GetIntParam (paramIndex), vmValType (VTP_INT, 1, 1, true), frames, size);

    // Convert to vector
    dest.clear ();
    for (int i = 0; i < size; i++)
        dest.push_back (frames [i]);

    return true;
}
void WrapNewSprite_3 (TomVM& vm) {

    // Allocate sprite and set an array of textures
    // Read textures
    TextureVec textures;
    if (!GetTextures (vm, 1, textures))
        return;

    // Allocate sprite
    glBasicSprite *sprite = NewSprite (vm);

    // Set textures
    sprite->SetTextures (textures);
    Redraw ();
}
void WrapNewTileMap (TomVM& vm) {

    // Allocate sprite
    NewTileMap (vm);
    Redraw ();
}
void WrapNewTileMap_2 (TomVM& vm) {

    // Allocate sprite, and set single texture
    NewTileMap (vm)->SetTexture (vm.GetIntParam(1));
    Redraw ();
}
void WrapNewTileMap_3 (TomVM& vm) {

    // Allocate sprite and set an array of textures
    // Read textures
    TextureVec textures;
    if (!GetTextures (vm, 1, textures))
        return;

    // Allocate sprite
    glBasicSprite *sprite = NewTileMap (vm);

    // Set textures
    sprite->SetTextures (textures);
    Redraw ();
}
inline bool IsBasicSprite (int index) {
    return sprites.IndexStored (index);
}
inline glBasicSprite* BasicSprite (int index) {
    assert (IsBasicSprite (index));
    return sprites.Value (index);
}
inline bool IsSprite (int index) {
    return IsBasicSprite (index)
        && BasicSprite (index)->Type() == SPR_SPRITE;
}
inline glSprite* Sprite (int index) {
    assert (IsSprite (index));
    return (glSprite *) BasicSprite (index);
}
inline bool IsTileMap (int index) {
    return IsBasicSprite (index)
        && BasicSprite (index)->Type () == SPR_TILEMAP;
}
inline glTileMap* TileMap (int index) {
    assert (IsTileMap (index));
    return (glTileMap *) BasicSprite (index);
}
void WrapDeleteSprite (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsBasicSprite (index)) {
        sprites.Free (index);
        Redraw ();
    }
}
void WrapBindSprite (TomVM& vm) {
    boundSprite = vm.GetIntParam (1);
}
void WrapSprSetTexture (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->SetTexture (vm.GetIntParam (1));
        Redraw ();
    }
}
void WrapSprSetTextures (TomVM& vm) {
    TextureVec textures;
    if (sprites.IndexStored (boundSprite) && GetTextures (vm, 1, textures)) {
        sprites.Value (boundSprite)->SetTextures (textures);
        Redraw ();
    }
}
void WrapSprAddTexture (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->AddTexture (vm.GetIntParam (1));
        Redraw ();
    }
}
void WrapSprAddTextures (TomVM& vm) {
    TextureVec textures;
    if (sprites.IndexStored (boundSprite) && GetTextures (vm, 1, textures)) {
        sprites.Value (boundSprite)->AddTextures (textures);
        Redraw ();
    }
}
void WrapSprSetFrame (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        Sprite (boundSprite)->SetFrame (vm.GetRealParam (1));
        Redraw ();
    }
}
void WrapSprSetX (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_x = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetY (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_y = vm.GetRealParam (1);
        Redraw ();
    }
}
float vec [4];
void ReadVec (TomVM& vm, int paramIndex) {

    // Read data
    vec [0] = 0;
    vec [1] = 0;
    vec [2] = 0;
    vec [3] = 1;
    ReadArray (vm.Data (), vm.GetIntParam (paramIndex), vmValType (VTP_REAL, 1, 1, true), vec, 4);
}
void WrapSprSetPos (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        ReadVec (vm, 1);
        sprites.Value (boundSprite)->m_x = vec [0];
        sprites.Value (boundSprite)->m_y = vec [1];
        Redraw ();
    }
}
void WrapSprSetPos_2 (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_x = vm.GetRealParam (2);
        sprites.Value (boundSprite)->m_y = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetZOrder (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->SetZOrder (vm.GetRealParam (1));
        Redraw ();
    }
}
void WrapSprSetXSize (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_xSize = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetYSize (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_ySize = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetSize (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        ReadVec (vm, 1);
        sprites.Value (boundSprite)->m_xSize = vec [0];
        sprites.Value (boundSprite)->m_ySize = vec [1];
        Redraw ();
    }
}
void WrapSprSetSize_2 (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_xSize = vm.GetRealParam (2);
        sprites.Value (boundSprite)->m_ySize = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetScale (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_scale = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetXCentre (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_xCentre = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetYCentre (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_yCentre = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetXFlip (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_xFlip = vm.GetIntParam (1) != 0;
        Redraw ();
    }
}
void WrapSprSetYFlip (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_yFlip = vm.GetIntParam (1) != 0;
        Redraw ();
    }
}
void WrapSprSetVisible (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_visible = vm.GetIntParam (1) != 0;
        Redraw ();
    }
}
void WrapSprSetAngle (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_angle = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetColour (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        ReadVec (vm, 1);
        int size = ArrayDimensionSize (vm.Data (), vm.GetIntParam (1), 0);
        if (size > 4)
            size = 4;
        memcpy (sprites.Value (boundSprite)->m_colour, vec, sizeof (vec [0]) * size);
        Redraw ();
    }
}
void WrapSprSetColour_2 (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_colour [0] = vm.GetRealParam (3);
        sprites.Value (boundSprite)->m_colour [1] = vm.GetRealParam (2);
        sprites.Value (boundSprite)->m_colour [2] = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetColour_3 (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_colour [0] = vm.GetRealParam (4);
        sprites.Value (boundSprite)->m_colour [1] = vm.GetRealParam (3);
        sprites.Value (boundSprite)->m_colour [2] = vm.GetRealParam (2);
        sprites.Value (boundSprite)->m_colour [3] = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetAlpha (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_colour [3] = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetParallax (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_parallax = vm.GetIntParam (1) != 0;
        Redraw ();
    }
}
void WrapSprSetSolid (TomVM& vm) {
    if (sprites.IndexStored (boundSprite)) {
        sprites.Value (boundSprite)->m_solid = vm.GetIntParam (1) != 0;
        Redraw ();
    }
}
void WrapResizeSpriteArea (TomVM& vm) {
    int width = vm.GetRealParam (2), height = vm.GetRealParam (1);
    if (width <= 0 || height <= 0) {
        vm.FunctionError ("Width and height must both be greater than 0");
        return;
    }
    ((glSpriteEngine *) appText)->SetWidth  (width);
    ((glSpriteEngine *) appText)->SetHeight (height);
}
void WrapSpriteAreaWidth (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->Width ();
}
void WrapSpriteAreaHeight (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->Height ();
}
void WrapSprFrame (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            Sprite (boundSprite)->Frame () : 0;
}
void WrapSprX (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_x : 0;
}
void WrapSprY (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_y : 0;
}
void WrapSprPos (TomVM& vm) {
    vmReal result [2] = {0, 0};
    if (sprites.IndexStored (boundSprite)) {
        result [0] = sprites.Value (boundSprite)->m_x;
        result [1] = sprites.Value (boundSprite)->m_y;
    }
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 2, result);
}
void WrapSprZOrder (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->ZOrder () : 0;
}
void WrapSprXSize (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_xSize : 0;
}
void WrapSprYSize (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_ySize : 0;
}
void WrapSprScale (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_scale : 0;
}
void WrapSprXCentre (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_xCentre : 0;
}
void WrapSprYCentre (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_yCentre : 0;
}
void WrapSprXFlip (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            (sprites.Value (boundSprite)->m_xFlip ? -1 : 0) : 0;
}
void WrapSprYFlip (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            (sprites.Value (boundSprite)->m_yFlip ? -1 : 0) : 0;
}
void WrapSprVisible (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            (sprites.Value (boundSprite)->m_visible ? -1 : 0) : 0;
}
void WrapSprAngle (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_angle : 0;
}
void WrapSprColour (TomVM& vm) {
    float result [4] = {0, 0, 0, 0};
    if (sprites.IndexStored (boundSprite))
        memcpy (result, sprites.Value (boundSprite)->m_colour, sizeof (result [0]) * 4);
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 4, result);
}
void WrapSprAlpha (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            sprites.Value (boundSprite)->m_colour [3] : 0;
}
void WrapSprParallax (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            (sprites.Value (boundSprite)->m_parallax ? -1 : 0) : 0;
}
void WrapSprSolid (TomVM& vm) {
    vm.Reg ().RealVal () =  sprites.IndexStored (boundSprite) ?
                            (sprites.Value (boundSprite)->m_solid ? -1 : 0) : 0;
}
void WrapSprLeft (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        glSprite *sprite = Sprite (boundSprite);
        vm.Reg ().RealVal () = sprite->m_x + -sprite->m_xCentre * (sprite->m_xSize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprRight (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        glSprite *sprite = Sprite (boundSprite);
        vm.Reg ().RealVal () = sprite->m_x + (1 - sprite->m_xCentre) * (sprite->m_xSize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprTop (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        glSprite *sprite = Sprite (boundSprite);
        vm.Reg ().RealVal () = sprite->m_y + -sprite->m_yCentre * (sprite->m_ySize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprBottom (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        glSprite *sprite = Sprite (boundSprite);
        vm.Reg ().RealVal () = sprite->m_y + (1 - sprite->m_yCentre) * (sprite->m_ySize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprXVel (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            Sprite (boundSprite)->m_xd : 0;
}
void WrapSprYVel (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            Sprite (boundSprite)->m_yd : 0;
}
void WrapSprVel (TomVM& vm) {
    vmReal result [2] = {0, 0};
    if (IsSprite (boundSprite)) {
        result [0] = Sprite (boundSprite)->m_xd;
        result [1] = Sprite (boundSprite)->m_yd;
    }
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 2, result);
}
void WrapSprSpin (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            Sprite (boundSprite)->m_angled : 0;
}
void WrapSprAnimSpeed (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            Sprite (boundSprite)->m_angled : 0;
}
void WrapSprAnimLoop (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            (Sprite (boundSprite)->m_animLoop ? -1 : 0) : 0;
}
void WrapSprAnimDone (TomVM& vm) {
    vm.Reg ().RealVal () =  IsSprite (boundSprite) ?
                            (Sprite (boundSprite)->AnimDone () ? -1 : 0) : 0;
}
void WrapSprFrame_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            Sprite (index)->Frame () : 0;
}
void WrapSprX_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_x : 0;
}
void WrapSprY_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_y : 0;
}
void WrapSprPos_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vmReal result [2] = {0, 0};
    if (sprites.IndexStored (index)) {
        result [0] = sprites.Value (index)->m_x;
        result [1] = sprites.Value (index)->m_y;
    }
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 2, result);
}
void WrapSprZOrder_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->ZOrder () : 0;
}
void WrapSprXSize_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_xSize : 0;
}
void WrapSprYSize_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_ySize : 0;
}
void WrapSprScale_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_scale : 0;
}
void WrapSprXCentre_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_xCentre : 0;
}
void WrapSprYCentre_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_yCentre : 0;
}
void WrapSprXFlip_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            (sprites.Value (index)->m_xFlip ? -1 : 0) : 0;
}
void WrapSprYFlip_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            (sprites.Value (index)->m_yFlip ? -1 : 0) : 0;
}
void WrapSprVisible_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            (sprites.Value (index)->m_visible ? -1 : 0) : 0;
}
void WrapSprAngle_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_angle : 0;
}
void WrapSprColour_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    float result [4] = {0, 0, 0, 0};
    if (sprites.IndexStored (index))
        memcpy (result, sprites.Value (index)->m_colour, sizeof (result [0]) * 4);
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 4, result);
}
void WrapSprAlpha_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            sprites.Value (index)->m_colour [3] : 0;
}
void WrapSprParallax_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            (sprites.Value (index)->m_parallax ? -1 : 0) : 0;
}
void WrapSprSolid_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  sprites.IndexStored (index) ?
                            (sprites.Value (index)->m_solid ? -1 : 0) : 0;
}
void WrapSprLeft_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsSprite (index)) {
        glSprite *sprite = Sprite (index);
        vm.Reg ().RealVal () = sprite->m_x + -sprite->m_xCentre * (sprite->m_xSize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprRight_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsSprite (index)) {
        glSprite *sprite = Sprite (index);
        vm.Reg ().RealVal () = sprite->m_x + (1 - sprite->m_xCentre) * (sprite->m_xSize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprTop_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsSprite (index)) {
        glSprite *sprite = Sprite (index);
        vm.Reg ().RealVal () = sprite->m_y + -sprite->m_yCentre * (sprite->m_ySize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprBottom_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsSprite (index)) {
        glSprite *sprite = Sprite (index);
        vm.Reg ().RealVal () = sprite->m_y + (1 - sprite->m_yCentre) * (sprite->m_ySize * sprite->m_scale);
    }
    else
        vm.Reg ().RealVal () = 0;
}
void WrapSprXVel_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            Sprite (index)->m_xd : 0;
}
void WrapSprYVel_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            Sprite (index)->m_yd : 0;
}
void WrapSprVel_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vmReal result [2] = {0, 0};
    if (IsSprite (index)) {
        result [0] = Sprite (index)->m_xd;
        result [1] = Sprite (index)->m_yd;
    }
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 2, result);
}
void WrapSprSpin_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            Sprite (index)->m_angled : 0;
}
void WrapSprAnimSpeed_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            Sprite (index)->m_angled : 0;
}
void WrapSprAnimLoop_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            (Sprite (index)->m_animLoop ? -1 : 0) : 0;
}
void WrapSprAnimDone_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().RealVal () =  IsSprite (index) ?
                            (Sprite (index)->AnimDone () ? -1 : 0) : 0;
}
void WrapSprSetXVel (TomVM& vm) {
    if (IsSprite (boundSprite))
        Sprite (boundSprite)->m_xd = vm.GetRealParam (1);
}
void WrapSprSetYVel (TomVM& vm) {
    if (IsSprite (boundSprite))
        Sprite (boundSprite)->m_yd = vm.GetRealParam (1);
}
void WrapSprSetVel (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        ReadVec (vm, 1);
        Sprite (boundSprite)->m_xd = vec [0];
        Sprite (boundSprite)->m_yd = vec [1];
        Redraw ();
    }
}
void WrapSprSetVel_2 (TomVM& vm) {
    if (IsSprite (boundSprite)) {
        Sprite (boundSprite)->m_xd = vm.GetRealParam (2);
        Sprite (boundSprite)->m_yd = vm.GetRealParam (1);
        Redraw ();
    }
}
void WrapSprSetSpin (TomVM& vm) {
    if (IsSprite (boundSprite))
        Sprite (boundSprite)->m_angled = vm.GetRealParam (1);
}
void WrapSprSetAnimSpeed (TomVM& vm) {
    if (IsSprite (boundSprite))
        Sprite (boundSprite)->m_framed = vm.GetRealParam (1);
}
void WrapSprSetAnimLoop (TomVM& vm) {
    if (IsSprite (boundSprite))
        Sprite (boundSprite)->m_animLoop = vm.GetIntParam (1) != 0;
}
void WrapAnimateSprites (TomVM& vm) {
    ((glSpriteEngine *) appText)->Animate ();
    Redraw ();
}
void WrapCopySprite (TomVM& vm) {

    // CopySprite (dest, source)
    // Copies sprite "source" to sprite "dest"
    // Sprites must be the same type. I.e. you can't copy a standard sprite
    // to a tile map or vice versa.
    int index = vm.GetIntParam (1);
    if (    IsBasicSprite (boundSprite) && IsBasicSprite (index)
        &&  BasicSprite (boundSprite)->Type () == BasicSprite (index)->Type ()) {
        if (IsSprite (boundSprite))
            Sprite (boundSprite)->Copy (*Sprite (index));
        else if (IsTileMap (boundSprite))
            TileMap (boundSprite)->Copy (*TileMap (index));
        Redraw ();
    }
}
void WrapSprType (TomVM& vm) {
    vm.Reg ().IntVal () =   IsBasicSprite (boundSprite) ?
                            BasicSprite (boundSprite)->Type () : SPR_INVALID;
}
void WrapSprXTiles (TomVM& vm) {
    vm.Reg ().IntVal () =   IsTileMap (boundSprite ) ?
                            TileMap (boundSprite)->XTiles () : 0;
}
void WrapSprYTiles (TomVM& vm) {
    vm.Reg ().IntVal () =   IsTileMap (boundSprite ) ?
                            TileMap (boundSprite)->YTiles () : 0;
}
void WrapSprType_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().IntVal () =   IsBasicSprite (index) ?
                            BasicSprite (index)->Type () : SPR_INVALID;
}
void WrapSprXTiles_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().IntVal () =   IsTileMap (index ) ?
                            TileMap (index)->XTiles () : 0;
}
void WrapSprYTiles_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    vm.Reg ().IntVal () =   IsTileMap (index ) ?
                            TileMap (index)->YTiles () : 0;
}
void GetTiles (TomVM& vm, int paramIndex, int& xSize, int& ySize, IntVec& dest) {

    // Read in texture array and convert to vector (for storage in sprite)
    int index = vm.GetIntParam (paramIndex);
    xSize = ArrayDimensionSize (vm.Data (), index, 0);
    ySize = ArrayDimensionSize (vm.Data (), index, 1);

    // Size must be valid and add up to 1 million or less tiles
    dest.clear ();
    if (xSize > 0 && ySize > 0) {

        // Read data into temp buffer
        int *buffer = new int [xSize * ySize];
        ReadArray (vm.Data (), index, vmValType (VTP_INT, 2, 1, true), buffer, xSize * ySize);

        // Convert to vector
        for (int i = 0; i < xSize * ySize; i++)
            dest.push_back (buffer [i]);

        // Free temp buffer
        delete[] buffer;
    }
}
void WrapSprSetTiles (TomVM& vm) {

    if (IsTileMap (boundSprite)) {

        // Get tiles from array param
        IntVec tiles;
        int xSize, ySize;
        GetTiles (vm, 1, xSize, ySize, tiles);

        // Set
        TileMap (boundSprite)->SetTiles (xSize, ySize, tiles);
        Redraw ();
    }
}
void WrapSprSetXRepeat (TomVM& vm) {
    if (IsTileMap (boundSprite)) {
        TileMap (boundSprite)->m_xRepeat = vm.GetIntParam (1) != 0 ? 1 : 0;
        Redraw ();
    }
}
void WrapSprSetYRepeat (TomVM& vm) {
    if (IsTileMap (boundSprite)) {
        TileMap (boundSprite)->m_yRepeat = vm.GetIntParam (1) != 0 ? 1 : 0;
        Redraw ();
    }
}
void WrapSprXRepeat (TomVM& vm) {
    if (IsTileMap (boundSprite))
        vm.Reg ().IntVal () = TileMap (boundSprite)->m_xRepeat ? -1 : 0;
    else
        vm.Reg ().IntVal () = 0;
}
void WrapSprYRepeat (TomVM& vm) {
    if (IsTileMap (boundSprite))
        vm.Reg ().IntVal () = TileMap (boundSprite)->m_yRepeat ? -1 : 0;
    else
        vm.Reg ().IntVal () = 0;
}
void WrapSprXRepeat_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsTileMap (index))
        vm.Reg ().IntVal () = TileMap (index)->m_xRepeat ? -1 : 0;
    else
        vm.Reg ().IntVal () = 0;
}
void WrapSprYRepeat_2 (TomVM& vm) {
    int index = vm.GetIntParam (1);
    if (IsTileMap (index))
        vm.Reg ().IntVal () = TileMap (index)->m_yRepeat ? -1 : 0;
    else
        vm.Reg ().IntVal () = 0;
}
void WrapSprCameraSetX (TomVM& vm) {
    ((glSpriteEngine *) appText)->m_camX = vm.GetRealParam (1);
    Redraw ();
}
void WrapSprCameraSetY (TomVM& vm) {
    ((glSpriteEngine *) appText)->m_camY = vm.GetRealParam (1);
    Redraw ();
}
void WrapSprCameraSetZ (TomVM& vm) {
    ((glSpriteEngine *) appText)->m_camZ = vm.GetRealParam (1);
    Redraw ();
}
void WrapSprCameraSetPos (TomVM& vm) {
    ((glSpriteEngine *) appText)->m_camX = vm.GetRealParam (2);
    ((glSpriteEngine *) appText)->m_camY = vm.GetRealParam (1);
    Redraw ();
}
void WrapSprCameraSetPos_2 (TomVM& vm) {
    ReadVec (vm, 1);
    ((glSpriteEngine *) appText)->m_camX = vec [0];
    ((glSpriteEngine *) appText)->m_camY = vec [1];
    Redraw ();
}
void WrapSprCameraSetAngle (TomVM& vm) {
    ((glSpriteEngine *) appText)->m_camAngle = vm.GetRealParam (1);
    Redraw ();
}
void WrapSprCameraX (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->m_camX;
}
void WrapSprCameraY (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->m_camY;
}
void WrapSprCameraZ (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->m_camZ;
}
void WrapSprCameraPos (TomVM& vm) {
    vmReal result [2] = {0, 0};
    result [0] = ((glSpriteEngine *) appText)->m_camX;
    result [1] = ((glSpriteEngine *) appText)->m_camY;
    vm.Reg ().IntVal () = FillTempRealArray (vm.Data (), vm.DataTypes (), 2, result);
}
void WrapSprCameraAngle (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->m_camAngle;
}
void WrapSprCameraFOV (TomVM& vm) {
    vm.Reg ().RealVal () = ((glSpriteEngine *) appText)->FOV ();
}
void WrapSprCameraSetFOV (TomVM& vm) {
    ((glSpriteEngine *) appText)->SetFOV (vm.GetRealParam (1));
    Redraw ();
}
void WrapClearLine (TomVM& vm) {
    appText->ClearLine ();
    Redraw ();
}

void WrapClearRegion (TomVM& vm) {
    appText->ClearRegion (vm.GetIntParam (4), vm.GetIntParam (3), vm.GetIntParam (2), vm.GetIntParam (1));
    Redraw ();
}

void WrapTextScroll (TomVM& vm) {
    vm.Reg ().IntVal () = appText->Scroll ();
}

void WrapSetTextScroll (TomVM& vm) {
    appText->Scroll () = vm.GetIntParam (1);
}

///////////////////////////////////////////////////////////////////////////////
// Intialisation

void InitTomTextBasicLib (TomBasicCompiler& comp, glWindow *win, glTextGrid *text) {

    // Register resources
    comp.VM().AddResources (sprites);

    // Store pointers to window and text objects
    appWindow   = win;
    appText     = text;

    // Register initialisation functions
    comp.VM().AddInitFunc (Init);

    // Register constants
    comp.AddConstant("TEXT_SIMPLE", TEXT_SIMPLE);
    comp.AddConstant("TEXT_BUFFERED", TEXT_BUFFERED);
    comp.AddConstant("TEXT_OVERLAID", TEXT_OVERLAID);
    comp.AddConstant("VK_LBUTTON", VK_LBUTTON);             // Virtual key codes
    comp.AddConstant("VK_RBUTTON", VK_RBUTTON);
    comp.AddConstant("VK_CANCEL", VK_CANCEL);
    comp.AddConstant("VK_MBUTTON", VK_MBUTTON);
//    comp.AddConstant("VK_XBUTTON1", VK_XBUTTON1);
//    comp.AddConstant("VK_XBUTTON2", VK_XBUTTON2);
    comp.AddConstant("VK_BACK", VK_BACK);
    comp.AddConstant("VK_TAB", VK_TAB);
    comp.AddConstant("VK_CLEAR", VK_CLEAR);
    comp.AddConstant("VK_RETURN", VK_RETURN);
    comp.AddConstant("VK_SHIFT", VK_SHIFT);
    comp.AddConstant("VK_CONTROL", VK_CONTROL);
    comp.AddConstant("VK_MENU", VK_MENU);
    comp.AddConstant("VK_PAUSE", VK_PAUSE);
    comp.AddConstant("VK_CAPITAL", VK_CAPITAL);
    comp.AddConstant("VK_KANA", VK_KANA);
    comp.AddConstant("VK_HANGEUL", VK_HANGEUL);
    comp.AddConstant("VK_HANGUL", VK_HANGUL);
    comp.AddConstant("VK_JUNJA", VK_JUNJA);
    comp.AddConstant("VK_FINAL", VK_FINAL);
    comp.AddConstant("VK_HANJA", VK_HANJA);
    comp.AddConstant("VK_KANJI", VK_KANJI);
    comp.AddConstant("VK_ESCAPE", VK_ESCAPE);
    comp.AddConstant("VK_CONVERT", VK_CONVERT);
    comp.AddConstant("VK_NONCONVERT", VK_NONCONVERT);
    comp.AddConstant("VK_ACCEPT", VK_ACCEPT);
    comp.AddConstant("VK_MODECHANGE", VK_MODECHANGE);
    comp.AddConstant("VK_SPACE", VK_SPACE);
    comp.AddConstant("VK_PRIOR", VK_PRIOR);
    comp.AddConstant("VK_NEXT", VK_NEXT);
    comp.AddConstant("VK_END", VK_END);
    comp.AddConstant("VK_HOME", VK_HOME);
    comp.AddConstant("VK_LEFT", VK_LEFT);
    comp.AddConstant("VK_UP", VK_UP);
    comp.AddConstant("VK_RIGHT", VK_RIGHT);
    comp.AddConstant("VK_DOWN", VK_DOWN);
    comp.AddConstant("VK_SELECT", VK_SELECT);
    comp.AddConstant("VK_PRINT", VK_PRINT);
    comp.AddConstant("VK_EXECUTE", VK_EXECUTE);
    comp.AddConstant("VK_SNAPSHOT", VK_SNAPSHOT);
    comp.AddConstant("VK_INSERT", VK_INSERT);
    comp.AddConstant("VK_DELETE", VK_DELETE);
    comp.AddConstant("VK_HELP", VK_HELP);
    comp.AddConstant("VK_LWIN", VK_LWIN);
    comp.AddConstant("VK_RWIN", VK_RWIN);
    comp.AddConstant("VK_APPS", VK_APPS);
//    comp.AddConstant("VK_SLEEP", VK_SLEEP);
    comp.AddConstant("VK_NUMPAD0", VK_NUMPAD0);
    comp.AddConstant("VK_NUMPAD1", VK_NUMPAD1);
    comp.AddConstant("VK_NUMPAD2", VK_NUMPAD2);
    comp.AddConstant("VK_NUMPAD3", VK_NUMPAD3);
    comp.AddConstant("VK_NUMPAD4", VK_NUMPAD4);
    comp.AddConstant("VK_NUMPAD5", VK_NUMPAD5);
    comp.AddConstant("VK_NUMPAD6", VK_NUMPAD6);
    comp.AddConstant("VK_NUMPAD7", VK_NUMPAD7);
    comp.AddConstant("VK_NUMPAD8", VK_NUMPAD8);
    comp.AddConstant("VK_NUMPAD9", VK_NUMPAD9);
    comp.AddConstant("VK_MULTIPLY", VK_MULTIPLY);
    comp.AddConstant("VK_ADD", VK_ADD);
    comp.AddConstant("VK_SEPARATOR", VK_SEPARATOR);
    comp.AddConstant("VK_SUBTRACT", VK_SUBTRACT);
    comp.AddConstant("VK_DECIMAL", VK_DECIMAL);
    comp.AddConstant("VK_DIVIDE", VK_DIVIDE);
    comp.AddConstant("VK_F1", VK_F1);
    comp.AddConstant("VK_F2", VK_F2);
    comp.AddConstant("VK_F3", VK_F3);
    comp.AddConstant("VK_F4", VK_F4);
    comp.AddConstant("VK_F5", VK_F5);
    comp.AddConstant("VK_F6", VK_F6);
    comp.AddConstant("VK_F7", VK_F7);
    comp.AddConstant("VK_F8", VK_F8);
    comp.AddConstant("VK_F9", VK_F9);
    comp.AddConstant("VK_F10", VK_F10);
    comp.AddConstant("VK_F11", VK_F11);
    comp.AddConstant("VK_F12", VK_F12);
    comp.AddConstant("VK_F13", VK_F13);
    comp.AddConstant("VK_F14", VK_F14);
    comp.AddConstant("VK_F15", VK_F15);
    comp.AddConstant("VK_F16", VK_F16);
    comp.AddConstant("VK_F17", VK_F17);
    comp.AddConstant("VK_F18", VK_F18);
    comp.AddConstant("VK_F19", VK_F19);
    comp.AddConstant("VK_F20", VK_F20);
    comp.AddConstant("VK_F21", VK_F21);
    comp.AddConstant("VK_F22", VK_F22);
    comp.AddConstant("VK_F23", VK_F23);
    comp.AddConstant("VK_F24", VK_F24);
    comp.AddConstant("VK_NUMLOCK", VK_NUMLOCK);
    comp.AddConstant("VK_SCROLL", VK_SCROLL);
//    comp.AddConstant("VK_OEM_NEC_EQUAL", VK_OEM_NEC_EQUAL);
    comp.AddConstant("VK_LSHIFT", VK_LSHIFT);
    comp.AddConstant("VK_RSHIFT", VK_RSHIFT);
    comp.AddConstant("VK_LCONTROL", VK_LCONTROL);
    comp.AddConstant("VK_RCONTROL", VK_RCONTROL);
    comp.AddConstant("VK_LMENU", VK_LMENU);
    comp.AddConstant("VK_RMENU", VK_RMENU);
//    comp.AddConstant("VK_BROWSER_BACK", VK_BROWSER_BACK);
//    comp.AddConstant("VK_BROWSER_FORWARD", VK_BROWSER_FORWARD);
//    comp.AddConstant("VK_BROWSER_REFRESH", VK_BROWSER_REFRESH);
//    comp.AddConstant("VK_BROWSER_STOP", VK_BROWSER_STOP);
//    comp.AddConstant("VK_BROWSER_SEARCH", VK_BROWSER_SEARCH);
//    comp.AddConstant("VK_BROWSER_FAVORITES", VK_BROWSER_FAVORITES);
//    comp.AddConstant("VK_BROWSER_HOME", VK_BROWSER_HOME);
//    comp.AddConstant("VK_VOLUME_MUTE", VK_VOLUME_MUTE);
//    comp.AddConstant("VK_VOLUME_DOWN", VK_VOLUME_DOWN);
//    comp.AddConstant("VK_VOLUME_UP", VK_VOLUME_UP);
//    comp.AddConstant("VK_MEDIA_NEXT_TRACK", VK_MEDIA_NEXT_TRACK);
//    comp.AddConstant("VK_MEDIA_PREV_TRACK", VK_MEDIA_PREV_TRACK);
//    comp.AddConstant("VK_MEDIA_STOP", VK_MEDIA_STOP);
//    comp.AddConstant("VK_MEDIA_PLAY_PAUSE", VK_MEDIA_PLAY_PAUSE);
//    comp.AddConstant("VK_LAUNCH_MAIL", VK_LAUNCH_MAIL);
//    comp.AddConstant("VK_LAUNCH_MEDIA_SELECT", VK_LAUNCH_MEDIA_SELECT);
//    comp.AddConstant("VK_LAUNCH_APP1", VK_LAUNCH_APP1);
//    comp.AddConstant("VK_LAUNCH_APP2", VK_LAUNCH_APP2);
    comp.AddConstant("VK_PROCESSKEY", VK_PROCESSKEY);
//    comp.AddConstant("VK_ICO_CLEAR", VK_ICO_CLEAR);
//    comp.AddConstant("VK_PACKET", VK_PACKET);
//    comp.AddConstant("VK_OEM_RESET", VK_OEM_RESET);
//    comp.AddConstant("VK_OEM_JUMP", VK_OEM_JUMP);
//    comp.AddConstant("VK_OEM_PA1", VK_OEM_PA1);
//    comp.AddConstant("VK_OEM_PA2", VK_OEM_PA2);
//    comp.AddConstant("VK_OEM_PA3", VK_OEM_PA3);
//    comp.AddConstant("VK_OEM_WSCTRL", VK_OEM_WSCTRL);
//    comp.AddConstant("VK_OEM_CUSEL", VK_OEM_CUSEL);
//    comp.AddConstant("VK_OEM_ATTN", VK_OEM_ATTN);
//    comp.AddConstant("VK_OEM_FINISH", VK_OEM_FINISH);
//    comp.AddConstant("VK_OEM_COPY", VK_OEM_COPY);
//    comp.AddConstant("VK_OEM_AUTO", VK_OEM_AUTO);
//    comp.AddConstant("VK_OEM_ENLW", VK_OEM_ENLW);
//    comp.AddConstant("VK_OEM_BACKTAB", VK_OEM_BACKTAB);
    comp.AddConstant("VK_ATTN", VK_ATTN);
    comp.AddConstant("VK_CRSEL", VK_CRSEL);
    comp.AddConstant("VK_EXSEL", VK_EXSEL);
    comp.AddConstant("VK_EREOF", VK_EREOF);
    comp.AddConstant("VK_PLAY", VK_PLAY);
    comp.AddConstant("VK_ZOOM", VK_ZOOM);
    comp.AddConstant("MOUSE_LBUTTON", 0);
    comp.AddConstant("MOUSE_RBUTTON", 1);
    comp.AddConstant("MOUSE_MBUTTON", 2);

    // Register functions
    comp.AddFunction ("cls",        WrapCls,        compParamTypeList (),                           false,  false,   VTP_INT);
    comp.AddFunction ("print",      WrapPrint,      compParamTypeList () << VTP_STRING,             false,  false,   VTP_INT);
    comp.AddFunction ("printr",     WrapPrintr,     compParamTypeList () << VTP_STRING,             false,  false,   VTP_INT);
    comp.AddFunction ("printr",     WrapPrintr_2,   compParamTypeList (),                           false,  false,   VTP_INT);
    comp.AddFunction ("locate",     WrapLocate,     compParamTypeList () << VTP_INT << VTP_INT,     false,  false,   VTP_INT);
    comp.AddFunction ("swapbuffers",WrapSwapBuffers,compParamTypeList (),                           true,   false,   VTP_INT);
    comp.AddFunction ("drawtext",   WrapDrawText,   compParamTypeList (),                           true,   false,   VTP_INT);
    comp.AddFunction ("textmode",   WrapTextMode,   compParamTypeList () << VTP_INT,                true,   false,   VTP_INT);    
    comp.AddFunction ("resizetext", WrapResizeText, compParamTypeList () << VTP_INT << VTP_INT,     true,   false,   VTP_INT);
    comp.AddFunction ("textrows",   WrapTextRows,   compParamTypeList (),                           true,   true,    VTP_INT);
    comp.AddFunction ("textcols",   WrapTextCols,   compParamTypeList (),                           true,   true,    VTP_INT);
    comp.AddFunction ("input$",     WrapInput,      compParamTypeList (),                           true,   true,    VTP_STRING, true);
    comp.AddFunction ("inkey$",     WrapInkey,      compParamTypeList (),                           true,   true,    VTP_STRING);
    comp.AddFunction ("inscankey",  WrapInScanKey,  compParamTypeList (),                           true,   true,    VTP_INT);
    comp.AddFunction ("clearkeys",  WrapClearKeys,  compParamTypeList (),                           true,   false,   VTP_INT);
    comp.AddFunction ("showcursor", WrapShowCursor, compParamTypeList (),                           true,   false,   VTP_INT);
    comp.AddFunction ("hidecursor", WrapHideCursor, compParamTypeList (),                           true,   false,   VTP_INT);
    comp.AddFunction ("keydown",    WrapKeyDown,    compParamTypeList () << VTP_STRING,             true,   true,    VTP_INT);
    comp.AddFunction ("scankeydown",WrapScanKeyDown,compParamTypeList () << VTP_INT,                true,   true,    VTP_INT);
    comp.AddFunction ("charat$",    WrapCharAt,     compParamTypeList () << VTP_INT << VTP_INT,     true,   true,    VTP_STRING);
    comp.AddFunction ("color",      WrapColour,     compParamTypeList () << VTP_INT << VTP_INT << VTP_INT, true, false, VTP_INT);
    comp.AddFunction ("font",       WrapFont,       compParamTypeList () << VTP_INT,                true,   false,  VTP_INT);
    comp.AddFunction ("defaultfont",WrapDefaultFont,compParamTypeList (),                           true,   true,   VTP_INT);
    comp.AddFunction ("mouse_x",     WrapMouseX,    compParamTypeList (),                           true,   true,   VTP_REAL);
    comp.AddFunction ("mouse_y",     WrapMouseY,    compParamTypeList (),                           true,   true,   VTP_REAL);
    comp.AddFunction ("mouse_xd",    WrapMouseXD,   compParamTypeList (),                           true,   true,   VTP_REAL);
    comp.AddFunction ("mouse_yd",    WrapMouseYD,   compParamTypeList (),                           true,   true,   VTP_REAL);
    comp.AddFunction ("mouse_button",WrapMouseButton,compParamTypeList () << VTP_INT,               true,   true,   VTP_INT);
    comp.AddFunction ("mouse_wheel", WrapMouseWheel,compParamTypeList (),                           true,   true,   VTP_INT);
    comp.AddFunction ("clearline",   WrapClearLine, compParamTypeList (),                           true,   false,  VTP_INT);
    comp.AddFunction ("clearregion", WrapClearRegion,compParamTypeList () << VTP_INT << VTP_INT << VTP_INT << VTP_INT, true, false, VTP_INT);
    comp.AddFunction ("textscroll",  WrapTextScroll,compParamTypeList (),                           true,   true,   VTP_INT);
    comp.AddFunction ("settextscroll",WrapSetTextScroll,compParamTypeList () << VTP_INT,            true,   false,  VTP_INT);

    comp.AddConstant ("SPR_INVALID",    SPR_INVALID);
    comp.AddConstant ("SPR_SPRITE",     SPR_SPRITE);
    comp.AddConstant ("SPR_TILEMAP",    SPR_TILEMAP);
    comp.AddFunction ("newsprite",      WrapNewSprite,      compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("newsprite",      WrapNewSprite_2,    compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("newsprite",      WrapNewSprite_3,    compParamTypeList () << vmValType (VTP_INT, 1, 1, true), true, true,  VTP_INT);
    comp.AddFunction ("newtilemap",     WrapNewTileMap,     compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("newtilemap",     WrapNewTileMap_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("newtilemap",     WrapNewTileMap_3,   compParamTypeList () << vmValType (VTP_INT, 1, 1, true), true, true,  VTP_INT);
    comp.AddFunction ("deletesprite",   WrapDeleteSprite,   compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("bindsprite",     WrapBindSprite,     compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsettexture",  WrapSprSetTexture,  compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsettextures", WrapSprSetTextures, compParamTypeList () << vmValType (VTP_INT, 1, 1, true), true, false, VTP_INT);
    comp.AddFunction ("spraddtexture",  WrapSprAddTexture,  compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("spraddtextures", WrapSprAddTextures, compParamTypeList () << vmValType (VTP_INT, 1, 1, true), true, false, VTP_INT);
    comp.AddFunction ("sprsetframe",    WrapSprSetFrame,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetx",        WrapSprSetX,        compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsety",        WrapSprSetY,        compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetzorder",   WrapSprSetZOrder,   compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetpos",      WrapSprSetPos,      compParamTypeList () << vmValType (VTP_REAL, 1, 1, true),true, false, VTP_INT);
    comp.AddFunction ("sprsetpos",      WrapSprSetPos_2,    compParamTypeList () << VTP_REAL << VTP_REAL,            true, false, VTP_INT);
    comp.AddFunction ("sprsetxsize",    WrapSprSetXSize,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetysize",    WrapSprSetYSize,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetsize",     WrapSprSetSize,     compParamTypeList () << vmValType (VTP_REAL, 1, 1, true),true, false, VTP_INT);
    comp.AddFunction ("sprsetsize",     WrapSprSetSize_2,   compParamTypeList () << VTP_REAL << VTP_REAL,            true, false, VTP_INT);
    comp.AddFunction ("sprsetscale",    WrapSprSetScale,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetxcentre",  WrapSprSetXCentre,  compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetycentre",  WrapSprSetYCentre,  compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetxflip",    WrapSprSetXFlip,    compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsetyflip",    WrapSprSetYFlip,    compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsetvisible",  WrapSprSetVisible,  compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsetangle",    WrapSprSetAngle,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetcolor",    WrapSprSetColour,   compParamTypeList () << vmValType (VTP_REAL, 1, 1, true),true, false, VTP_INT);
    comp.AddFunction ("sprsetcolor",    WrapSprSetColour_2, compParamTypeList () << VTP_REAL << VTP_REAL << VTP_REAL,true, false, VTP_INT);
    comp.AddFunction ("sprsetcolor",    WrapSprSetColour_3, compParamTypeList () << VTP_REAL << VTP_REAL << VTP_REAL << VTP_REAL,true, false, VTP_INT);
    comp.AddFunction ("sprsetalpha",    WrapSprSetAlpha,    compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetparallax", WrapSprSetParallax, compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsetsolid",    WrapSprSetSolid,    compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("resizespritearea",   WrapResizeSpriteArea,   compParamTypeList () << VTP_REAL << VTP_REAL,                       true, false, VTP_INT);
    comp.AddFunction ("spriteareawidth",    WrapSpriteAreaWidth,    compParamTypeList (),                                               true, true,  VTP_REAL);
    comp.AddFunction ("spriteareaheight",   WrapSpriteAreaHeight,   compParamTypeList (),                                               true, true,  VTP_REAL);
    comp.AddFunction ("sprframe",       WrapSprFrame,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprx",           WrapSprX,           compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("spry",           WrapSprY,           compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprpos",         WrapSprPos,         compParamTypeList (),                                    true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("sprzorder",      WrapSprZOrder,      compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprxsize",       WrapSprXSize,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprysize",       WrapSprYSize,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprscale",       WrapSprScale,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprxcentre",     WrapSprXCentre,     compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprycentre",     WrapSprYCentre,     compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprxflip",       WrapSprXFlip,       compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("spryflip",       WrapSprYFlip,       compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprvisible",     WrapSprVisible,     compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprangle",       WrapSprAngle,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprcolor",       WrapSprColour,      compParamTypeList (),                                    true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("spralpha",       WrapSprAlpha,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprparallax",    WrapSprParallax,    compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprsolid",       WrapSprSolid,       compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprleft",        WrapSprLeft,        compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprright",       WrapSprRight,       compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprtop",         WrapSprTop,         compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprbottom",      WrapSprBottom,      compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprxvel",        WrapSprXVel,        compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("spryvel",        WrapSprYVel,        compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("sprvel",         WrapSprVel,         compParamTypeList (),                                    true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("sprspin",        WrapSprSpin,        compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("spranimspeed",   WrapSprAnimSpeed,   compParamTypeList (),                                    true, true,  VTP_REAL);
    comp.AddFunction ("spranimloop",    WrapSprAnimLoop,    compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("spranimdone",    WrapSprAnimDone,    compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprframe",       WrapSprFrame_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprx",           WrapSprX_2,         compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("spry",           WrapSprY_2,         compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprpos",         WrapSprPos_2,       compParamTypeList () << VTP_INT,                         true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("sprzorder",      WrapSprZOrder_2,    compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprxsize",       WrapSprXSize_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprysize",       WrapSprYSize_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprscale",       WrapSprScale_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprxcentre",     WrapSprXCentre_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprycentre",     WrapSprYCentre_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprxflip",       WrapSprXFlip_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("spryflip",       WrapSprYFlip_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprvisible",     WrapSprVisible_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprangle",       WrapSprAngle_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprcolor",       WrapSprColour_2,    compParamTypeList () << VTP_INT,                         true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("spralpha",       WrapSprAlpha_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprparallax",    WrapSprParallax_2,  compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprsolid",       WrapSprSolid_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprleft",        WrapSprLeft_2,      compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprright",       WrapSprRight_2,     compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprtop",         WrapSprTop_2,       compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprbottom",      WrapSprBottom_2,    compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprxvel",        WrapSprXVel_2,      compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("spryvel",        WrapSprYVel_2,      compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("sprvel",         WrapSprVel_2,       compParamTypeList () << VTP_INT,                         true, true,  vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("sprspin",        WrapSprSpin_2,      compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("spranimspeed",   WrapSprAnimSpeed_2, compParamTypeList () << VTP_INT,                         true, true,  VTP_REAL);
    comp.AddFunction ("spranimloop",    WrapSprAnimLoop_2,  compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("spranimdone",    WrapSprAnimDone_2,  compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprsetxvel",     WrapSprSetXVel,     compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetyvel",     WrapSprSetYVel,     compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetvel",      WrapSprSetVel,      compParamTypeList () << vmValType (VTP_REAL, 1, 1, true),true, false, VTP_INT);
    comp.AddFunction ("sprsetvel",      WrapSprSetVel_2,    compParamTypeList () << VTP_REAL << VTP_REAL,            true, false, VTP_INT);
    comp.AddFunction ("sprsetspin",     WrapSprSetSpin,     compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetanimspeed",WrapSprSetAnimSpeed,compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprsetanimloop", WrapSprSetAnimLoop, compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("animatesprites", WrapAnimateSprites, compParamTypeList (),                                    true, false, VTP_INT);
    comp.AddFunction ("copysprite",     WrapCopySprite,     compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprtype",        WrapSprType,        compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprxtiles",      WrapSprXTiles,      compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprytiles",      WrapSprYTiles,      compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprtype",        WrapSprType_2,      compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprxtiles",      WrapSprXTiles_2,    compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprytiles",      WrapSprYTiles_2,    compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprsettiles",    WrapSprSetTiles,    compParamTypeList () << vmValType (VTP_INT, 2, 1, true), true, false, VTP_INT);
    comp.AddFunction ("sprsetxrepeat",  WrapSprSetXRepeat,  compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprsetyrepeat",  WrapSprSetYRepeat,  compParamTypeList () << VTP_INT,                         true, false, VTP_INT);
    comp.AddFunction ("sprxrepeat",     WrapSprXRepeat,     compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("spryrepeat",     WrapSprYRepeat,     compParamTypeList (),                                    true, true,  VTP_INT);
    comp.AddFunction ("sprxrepeat",     WrapSprXRepeat_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("spryrepeat",     WrapSprYRepeat_2,   compParamTypeList () << VTP_INT,                         true, true,  VTP_INT);
    comp.AddFunction ("sprcamerasetx",  WrapSprCameraSetX,  compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprcamerasety",  WrapSprCameraSetY,  compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprcamerasetz",  WrapSprCameraSetZ,  compParamTypeList () << VTP_REAL,                        true, false, VTP_INT);
    comp.AddFunction ("sprcamerasetpos",WrapSprCameraSetPos,compParamTypeList () << VTP_REAL << VTP_REAL,            true, false, VTP_INT);
    comp.AddFunction ("sprcamerasetpos",WrapSprCameraSetPos_2,compParamTypeList () << vmValType (VTP_REAL, 1, 1, true),true, false, VTP_INT);
    comp.AddFunction ("sprcamerasetangle",WrapSprCameraSetAngle,compParamTypeList () << VTP_REAL,                    true, false, VTP_INT);
    comp.AddFunction ("sprcamerax",     WrapSprCameraX,     compParamTypeList (),                                    true, true, VTP_REAL);
    comp.AddFunction ("sprcameray",     WrapSprCameraY,     compParamTypeList (),                                    true, true, VTP_REAL);
    comp.AddFunction ("sprcameraz",     WrapSprCameraZ,     compParamTypeList (),                                    true, true, VTP_REAL);
    comp.AddFunction ("sprcamerapos",   WrapSprCameraPos,   compParamTypeList (),                                    true, true, vmValType (VTP_REAL, 1, 1, true));
    comp.AddFunction ("sprcameraangle", WrapSprCameraAngle, compParamTypeList (),                                    true, true, VTP_REAL);
    comp.AddFunction ("sprcamerafov",   WrapSprCameraFOV,   compParamTypeList (),                                    true, true, VTP_REAL);
    comp.AddFunction ("sprcamerasetfov",WrapSprCameraSetFOV,compParamTypeList () << VTP_REAL,                        true, false,VTP_INT);
}
